#
# Copyright (c) 2017-2019 TBD. All rights reserved.
# This file is part of TBD (see TBD).
# See the LICENSE file for licensing terms (TBD).
#

# TODO transfer the documentation strings to flex creators/instances

import numpy as np
import torch
from torch import autograd, nn
from torch.autograd import Variable
import warnings
from functools import wraps

from . import helpers, layers

verbose = False

class Flex(nn.Module):
    def __init__(self, creator):
        super(Flex, self).__init__()
        self.creator = creator
        self.layer = None

    def forward(self, *args):
        if self.layer is None:
            self.layer = self.creator(*args)
            if verbose: print("# created", self.layer)
        return self.layer.forward(*args)

    def __repr__(self):
        return "Flex:"+repr(self.layer)

    def __str__(self):
        return "Flex:"+str(self.layer)


def Linear(*args, **kw):
    def creator(x):
        assert x.ndimension() == 2
        return nn.Linear(x.size(1), *args, **kw)
    return Flex(creator)


def Conv1d(*args, **kw):
    def creator(x):
        assert x.ndimension() == 3
        d = x.size(1)
        return nn.Conv1d(x.size(1), *args, **kw)
    return Flex(creator)


def Conv2d(*args, **kw):
    def creator(x):
        assert x.ndimension() == 4
        return nn.Conv2d(x.size(1), *args, **kw)
    return Flex(creator)


def Conv3d(*args, **kw):
    def creator(x):
        assert x.ndimension() == 5
        return nn.Conv3d(x.size(1), *args, **kw)
    return Flex(creator)


def ConvTranspose1d(*args, **kw):
    def creator(x):
        assert x.ndimension() == 3
        d = x.size(1)
        return nn.ConvTranspose1d(x.size(1), *args, **kw)
    return Flex(creator)


def ConvTranspose2d(*args, **kw):
    def creator(x):
        assert x.ndimension() == 4
        return nn.ConvTranspose2d(x.size(1), *args, **kw)
    return Flex(creator)


def ConvTranspose3d(*args, **kw):
    def creator(x):
        assert x.ndimension() == 5
        return nn.ConvTranspose3d(x.size(1), *args, **kw)
    return Flex(creator)


def LSTM(*args, **kw):
    def creator(x):
        assert x.ndimension() == 3
        return layers.LSTM(x.size(2), *args, **kw)
    return Flex(creator)

def BDL_LSTM(*args, **kw):
    def creator(x):
        assert x.ndimension() == 3
        return layers.BDL_LSTM(x.size(1), *args, **kw)
    return Flex(creator)

Lstm1 = BDL_LSTM
Lstm1d = BDL_LSTM

def BDHW_LSTM(*args, **kw):
    def creator(x):
        assert x.ndimension() == 4
        return layers.BDHW_LSTM(x.size(1), *args, **kw)
    return Flex(creator)

Lstm2 = BDHW_LSTM
Lstm2d = BDHW_LSTM

def BatchNorm(*args, **kw):
    def creator(x):
        assert x.ndimension() == 2
        assert x.size(0) > 1, f"batch norm requires batch size > 1, got {x.shape}"
        return nn.BatchNorm1d(x.size(1), *args, **kw)
    return Flex(creator)

def BatchNorm1d(*args, **kw):
    def creator(x):
        assert x.ndimension() == 3
        return nn.BatchNorm1d(x.size(1), *args, **kw)
    return Flex(creator)


def BatchNorm2d(*args, **kw):
    def creator(x):
        assert x.ndimension() == 4
        return nn.BatchNorm2d(x.size(1), *args, **kw)
    return Flex(creator)


def BatchNorm3d(*args, **kw):
    def creator(x):
        assert x.ndimension() == 5
        return nn.BatchNorm3d(x.size(1), *args, **kw)
    return Flex(creator)


def replace_modules(model, f):
    for key in list(model._modules.keys()):
        sub = model._modules[key]
        replacement = f(sub)
        if replacement is not None:
            model._modules[key] = replacement
        else:
            replace_modules(sub, f)


def flex_replacer(module):
    if isinstance(module, Flex):
        return module.layer
    else:
        return None

def flex_freeze(model):
    replace_modules(model, flex_replacer)

def freeze(model):
    replace_modules(model, flex_replacer)

def shape_inference(model, tensor, dtype=None):
    if isinstance(tensor, (tuple, list)):
        tensor = torch.zeros(tensor, dtype=dtype)
    model.eval()
    with autograd.no_grad():
        model(tensor)
    replace_modules(model, flex_replacer)

def delete_modules(model, f):
    for key in list(model._modules.keys()):
        if f(model._modules[key]):
            del model._modules[key]



#
# Copyright (c) 2017-2019 TBD. All rights reserved.
# This file is part of TBD (see TBD).
# See the LICENSE file for licensing terms (TBD).
#
from __future__ import unicode_literals

import pdb
from builtins import next
import sys
import os
import numpy as np

import torch
from torch import nn
from torchmore import layers

def test_weighted_grad():
    pass

def test_Fun():
    mod = layers.Fun("lambda x: x[0]**2")
    a = torch.ones((2, 3, 4))
    b = mod(a)
    assert (b==1).all()
    assert tuple(b.shape) == (3, 4)

def test_Info():
    mod = layers.Info()
    a = torch.ones((2, 3, 4))
    b = mod(a)
    assert (b==a).all()

def test_CheckSizes():
    mod = layers.CheckSizes(2, 3, 4)
    a = torch.ones((2, 3, 4))
    b = mod(a)
    assert (b==a).all()

def test_Device():
    mod = layers.Device("cpu")
    a = torch.ones((2, 3, 4))
    b = mod(a)
    assert (b.cpu()==a.cpu()).all()

def test_CheckRange():
    mod = layers.CheckRange(0, 2)
    a = torch.ones((2, 3, 4))
    b = mod(a)
    assert (b.cpu()==a.cpu()).all()

def test_Input():
    mod = layers.Input("ABCD", sizes=(4, 5, 2, 3))
    a = torch.ones((2, 3, 4, 5))
    a.order = "CDAB"
    b = mod(a)
    assert tuple(b.shape) == (4, 5, 2, 3)

def test_Reorder():
    mod = layers.Reorder("ABC", "CBA")
    a = torch.ones((2, 3, 4))
    b = mod(a)
    assert tuple(b.shape) == (4, 3, 2)

def test_Permute():
    mod = layers.Permute(2, 1, 0)
    a = torch.ones((2, 3, 4))
    b = mod(a)
    assert tuple(b.shape) == (4, 3, 2)

def test_Reshape():
    mod = layers.Reshape(0, (1, 2))
    a = torch.ones((2, 3, 4))
    b = mod(a)
    assert tuple(b.shape) == (2, 12)

def test_Viewer():
    mod = layers.Viewer(6, 4)
    a = torch.ones((2, 3, 4))
    b = mod(a)
    assert tuple(b.shape) == (6, 4)

def test_BDL_LSTM():
    mod = layers.BDL_LSTM(3, 20)
    a = torch.ones((17, 3, 100))
    b = mod(a)
    assert tuple(b.shape) == (17, 20, 100)

def test_BDHW_LSTM():
    mod = layers.BDHW_LSTM(3, 20)
    a = torch.ones((17, 3, 48, 64))
    b = mod(a)
    assert tuple(b.shape) == (17, 40, 48, 64)

def test_BDHW_LSTM_to_BDH():
    mod = layers.BDHW_LSTM_to_BDH(3, 13)
    a = torch.ones((17, 3, 48, 64))
    b = mod(a)
    assert tuple(b.shape) == (17, 13, 48)

def test_NoopSub():
    mod = layers.NoopSub(sub=nn.Identity())
    a = torch.ones((17, 11, 64, 64))
    b = mod(a)
    assert tuple(b.size()) == (17, 11, 64, 64)

def test_KeepSize():
    mod = layers.KeepSize(sub=nn.FractionalMaxPool2d((3, 3), output_ratio=(0.7, 0.9)))
    a = torch.ones((17, 11, 64, 64))
    b = mod(a)
    assert tuple(b.size()) == (17, 11, 64, 64)

def test_Additive():
    mod = layers.Additive(
        nn.Conv2d(11, 17, 3, padding=1),
        nn.Conv2d(11, 17, 5, padding=2)
    )
    a = torch.ones((17, 11, 64, 64))
    b = mod(a)
    assert tuple(b.size()) == (17, 17, 64, 64)

def test_Parallel():
    mod = layers.Parallel(
        nn.Conv2d(1, 4, 3, padding=(1, 1)),
        nn.Conv2d(1, 7, 3, padding=(1, 1)))
    a = torch.ones((17, 1, 64, 64))
    b = mod(a)
    assert tuple(b.size()) == (17, 11, 64, 64)

def test_SimplePooling2d():
    mod = layers.SimplePooling2d([nn.Conv2d(1, 1, 3, padding=(1, 1))])
    a = torch.ones((17, 1, 64, 64))
    b = mod(a)
    assert a.size() == b.size()

def test_AcrossPooling2d():
    mod = layers.AcrossPooling2d(
        [nn.Conv2d(1, 1, 3, padding=(1, 1))],
        [nn.Conv2d(1, 7, 3, padding=(1, 1))])
    a = torch.ones((17, 1, 64, 64))
    b = mod(a)
    assert tuple(b.size()) == (17, 8, 64, 64)




import sys
import numpy as np
import torch
from torch import autograd, nn
from torch.nn import functional as F
from . import layers, flex

def fc_block(sizes, batchnorm=True, nonlin=nn.ReLU, flatten=True):
    assert len(sizes) >= 1, sizes
    result = [nn.Flatten()]
    for i, s in enumerate(sizes[:-1]):
        result += [flex.Linear(s)]
        if batchnorm:
            result += [flex.BatchNorm()]
        result += [nonlin()]
    result += [flex.Linear(sizes[-1])]
    return result

def conv2d_block(d, r=3, mp=None, fmp=None, repeat=1, batchnorm=True, nonlin=nn.ReLU):
    """Generate a conv layer with batchnorm and optional maxpool."""
    result = []
    for i in range(repeat):
        result += [flex.Conv2d(d, r, padding=(r//2, r//2))]
        if batchnorm:
            result += [flex.BatchNorm2d()]
        result += [nonlin()]
    if fmp is not None:
        assert mp is None, (fmp, mp)
        result += [nn.FractionalMaxPool2d(3, output_ratio=fmp)]
    elif mp is not None:
        result += [nn.MaxPool2d(mp)]
    return result

class UnetLayer(nn.Module):
    """Resolution pyramid layer using convolutions and upscaling.
    """
    def __init__(self, d, r=3, sub=None, post=None):
        super().__init__()
        self.down = nn.MaxPool2d(2)
        self.up = flex.ConvTranspose2d(d, r, stride=2, padding=1, output_padding=1)
        if isinstance(sub, list):
            sub = nn.Sequential(*sub)
        self.sub = sub
        self.post = post
    def forward(self, x):
        b, d, h, w = x.size()
        assert h%2==0 and w%2==0, x.size()
        lo = self.down(x)
        lo1 = self.sub(lo)
        hi = self.up(lo1)
        result = torch.cat([x, hi], dim=1)
        if self.post is not None:
            result = self.post(result)
        return result

def ResnetBottleneck(d,  b, r=3, identity=None, post=None):
    return layers.Additive(
        identity or nn.Identity(),
        nn.Sequential(
            nn.Conv2d(d, b, 1),
            nn.BatchNorm2d(b),
            nn.ReLU(),
            nn.Conv2d(b, b, r, padding=r//2),
            nn.BatchNorm2d(b),
            nn.ReLU(),
            nn.Conv2d(b, d, 1)
        ),
        post = post or nn.BatchNorm2d(d)
    )

def ResnetBlock(d, r=3, identity=None, post=None):
    return layers.Additive(
        identity or nn.Identity(),
        nn.Sequential(
            nn.Conv2d(d, d, r, padding=r//2),
            nn.BatchNorm2d(d),
            nn.ReLU(),
            nn.Conv2d(d, d, r, padding=r//2),
            nn.BatchNorm2d(d)
        ),
        post = post or nn.BatchNorm2d(d)
    )

def resnet_blocks(n, d, r=3):
    return [ResnetBlock(d, r) for _ in range(n)]

def make_unet(sizes, r=3, repeat=3, sub=None):
    if len(sizes)==1:
        if sub is None:
            return nn.Sequential(*conv2d_block(sizes[0], r, repeat=repeat))
        else:
            return UnetLayer(sizes[0], sub=sub)
    else:
        return UnetLayer(sizes[0],
            sub=make_unet(sizes[1:], r=r, repeat=repeat, sub=sub)
        )
    




#
# Copyright (c) 2017-2019 TBD. All rights reserved.
# This file is part of TBD (see TBD).
# See the LICENSE file for licensing terms (TBD).
#

import inspect


class LocalImport(object):

    def __init__(self, names):
        if not isinstance(names, dict):
            names = vars(names)
        self.names = names

    def __enter__(self):
        self.frame = inspect.currentframe()
        bindings = self.frame.f_back.f_globals
        self.old_bindings = {k: bindings.get(
            k, None) for k in list(self.names.keys())}
        bindings.update(self.names)

    def __exit__(self, some_type, value, traceback):
        del some_type, value, traceback
        bindings = self.frame.f_back.f_globals
        bindings.update(self.old_bindings)
        extras = [k for k, v in list(self.old_bindings.items()) if v is None]
        for k in extras:
            del bindings[k]
        del self.frame



#
# Copyright (c) 2017-2019 TBD. All rights reserved.
# This file is part of TBD (see TBD).
# See the LICENSE file for licensing terms (TBD).
#

import sys
import numpy as np
import torch
from torch import autograd, nn
from torch.nn import functional as F

def deprecated(f):
    def g(*args, **kw):
        raise Exception("deprecated")
    return g


def conform1(a, *args, slop=None, dims=True):
    """Trim the remaining args down to the size of the first."""
    if dims is True: dims = arange(a.ndimension())
    if slop is not None:
        # FIXME convert to torch
        target = np.array(a.shape)
        sizes = np.array([a.shape for a in args])
        deltas = np.amax(np.abs(sizes-target[np.newaxis,:]), 0)
        for i in dims:
            assert deltas[i] <= slop, (sizes, deltas)
    box = tuple(slice(j) if i in dims else slice(None)
                for i,j in enumerate(a.shape))
    return tuple([a[box]] + [arg[box] for arg in args])


def conform(*args, slop=None, dims=True):
    """Trim all args to the size of the smallest arg."""
    if slop is not None:
        # FIXME convert to torch
        sizes = np.array([a.shape for a in args])
        deltas = np.amax(sizes, 0)-np.amin(sizes, 0)
        for i in dims:
            assert deltas[i] <= slop, (sizes, deltas)
    box = map(min, zip(*[a.shape for a in args]))
    box = tuple(slice(j) if i in dims else slice(None)
                for i,j in enumerate(box))
    return tuple([arg[box] for arg in args])


def reorder(x, old, new, set_order=True):
    """Reorder dimensions according to strings.

    E.g., reorder(x, "BLD", "LBD")
    """
    assert isinstance(old, str) and isinstance(new, str)
    assert set(old)==set(new) and len(old)==len(new) and len(set(old))==len(old), (old, new)
    permutation = tuple([old.find(c) for c in new])
    assert len(old) == x.ndimension(), (old, x.size())
    result = x.permute(permutation).contiguous()
    if set_order:
        result.order = new
    return result

def check_order(x, order):
    if hasattr(x, "order"):
        if x.order != order:
            raise ValueError(f"expected order {order}, got {x.order}")

class WeightedGrad(autograd.Function):
    """Reweight the gradient using the given weights."""

    def forward(self, input, weights):
        self.weights = weights
        return input

    def backward(self, grad_output):
        return grad_output * self.weights, None

def weighted_grad(x, y):
    return WeightedGrad()(x, y)


class Fun(nn.Module):
    """Turn an arbitrary function into a layer."""

    def __init__(self, f, info=None):
        super().__init__()
        assert isinstance(f, str), type(f)
        self.f = eval(f)
        self.f_str = f
        self.info = info

    def __getnewargs__(self):
        return (self.f_str, self.info)

    def forward(self, x):
        return self.f(x)

    def __repr__(self):
        return "Fun {} {}".format(self.info, self.f_str)

class Fun_(nn.Module):
    """Turn an arbitrary function into a layer."""

    def __init__(self, f, info=None):
        super().__init__()
        assert callable(f)
        self.f = f
        self.info = info

    def forward(self, x):
        return self.f(x)

    def __repr__(self):
        return "Fun {} {}".format(self.info, self.f)



class Info(nn.Module):
    """Output information for the given input."""

    def __init__(self, info="", every=1000000):
        super().__init__()
        self.info = info
        self.count = 0
        self.every = every

    def forward(self, x):
        if self.count % self.every == 0:
            print(("Info", self.info, x.size(),
                  x.min().item(), x.max().item()))
        self.count += 1
        return x

    def __repr__(self):
        return "Info {} ({} % {})".format(self.info, self.count, self.every)


class CheckSizes(nn.Module):
    """Check tensor sizes against the given sizes.

    Specify ranges as integers (exact), tuples (low, high), or -1 (no check).
    """

    def __init__(self, *args, **kw):
        super().__init__()
        self.order = kw.get("order")
        self.name = kw.get("name")
        self.limits = [(x, x) if isinstance(x, int) else x for x in args]

    def forward(self, x):
        for (i, actual), (lo, hi) in zip(enumerate(tuple(x.size())), self.limits):
            if lo >= 0 and actual < lo:
                raise Exception("{} ({}): index {} too low ({} not >= {})"
                                .format(self.name, self.order,
                                        i, actual, lo))
            if hi >= 0 and actual > hi:
                raise Exception("{} ({}): index {} too high ({} not <= {})"
                                .format(self.name, self.order,
                                        i, actual, hi))
        return x

    def __repr__(self):
        return "CheckSizes({})".format(
            ", ".join([repr(x) for x in self.limits]))


class Device(nn.Module):
    """Always transfer tensor to device."""

    def __init__(self, device="cuda"):
        super().__init__()
        self.device = device

    def forward(self, x):
        return x.to(self.device)

    def __repr__(self):
        return "Device {}".format(self.device)


class CheckRange(nn.Module):
    """Check that values are within the given range."""

    def __init__(self, lo=-1e5, hi=1e5, name=""):
        super().__init__()
        self.valid = (lo, hi)

    def forward(self, x):
        assert x.min().item() >= self.valid[0], \
            (self.name, data(x).min(), self.valid)
        assert x.max().item() <= self.valid[1], \
            (self.name, data(x).max(), self.valid)
        return x

    def __repr__(self):
        return "CheckRange({}, {}, name=\"{}\")".format(
                           self.valid[0],
                           self.valid[1],
                           self.name)


class Input(nn.Module):
    def __init__(self, assume, reorder=None, range=None, sizes=None, device=True, dtype=torch.float32):
        """Declares the input for a network.

        :param order: order of axes (e.g., BDL, BHWD, etc.)
        :param dtype: dtype to convert to
        :param range: tuple giving low/high values
        :param device: input device to move to (True = auto)
        :param assume: default input order (when tensor doesn't have order attribute; None=required)
        """
        super().__init__()
        self.assume = assume
        self.reorder = reorder if reorder is not None else assume
        self.dtype = dtype
        self.range = range
        self.device = device
        self.param = torch.nn.Parameter(torch.zeros(1))
        self.sizes = sizes
    def forward(self, x):
        if self.range is not None:
            lo = x.min().item()
            hi = x.max().item()
            assert lo >= self.range[0] and hi <= self.range[1], (lo, hi, self.range)
        if self.reorder is not None:
            if hasattr(x, "order"):
                x = reorder(x, x.order, self.reorder)
            else:
                if self.assume is True or self.assume==self.reorder:
                    pass
                elif self.assume is None:
                    raise ValueError("input is required to have a .order property")
                else:
                    x = reorder(x, self.assume, self.reorder)
        if self.sizes is not None:
            for i, size in enumerate(self.sizes):
                if size is None:
                    continue
                elif isinstance(size, int):
                    assert x.size(i) == size, (i, x.size(i))
                elif isinstance(size, (list, tuple)):
                    lo, hi = size
                    assert x.size(i) >= lo and x.size(i) <= hi, (i, x.size(i), (lo, hi))
                else:
                    raise ValueError("bad size spec")
        if self.device is True:
            x = x.to(device=self.param.device, dtype=self.dtype)
        else:
            x = x.type(self.dtype)
        return x
    def __repr__(self):
        autodev = self.param.device if self.device else None
        return f"Input({self.assume}->{self.reorder} " + \
            f"{self.dtype} {self.range} {autodev} {self.sizes})"


class Reorder(nn.Module):
    """Reorder the dimensions using the given arguments."""

    def __init__(self, old, new):
        self.old = old
        self.new = new
        super().__init__()
        assert isinstance(old, str), old
        assert isinstance(new, str), new
        assert set(old) == set(new), (old, new)
        self.permutation = tuple([old.find(c) for c in new])

    def forward(self, x):
        return x.permute(*self.permutation).contiguous()

    def __repr__(self):
        return 'Reorder("{}", "{}")'.format(self.old, self.new)

class CheckOrder(nn.Module):
    def __init__(self, order=None):
        super().__init__()
        self.order = order
    def forward(self, x):
        if self.order is not None:
            check_order(x, self.order)
        return x

class Permute(nn.Module):
    """Permute the dimensions of the input tensor."""

    def __init__(self, *args):
        super().__init__()
        self.permutation = args

    def forward(self, x):
        return x.permute(*self.permutation).contiguous()

    def __repr__(self):
        return "Permute({})".format(", ".join(list(self.permutation)))


class Reshape(nn.Module):
    """Reshape an input tensor.

    Shapes can be specified as a list of integers and tuples.
    If specified as tuple, the corresponding input dimensions
    are collapsed into a single output dimension.
    """

    def __init__(self, *args):
        super().__init__()
        self.shape = args

    def forward(self, x):
        newshape = []
        for s in self.shape:
            if isinstance(s, int):
                newshape.append(int(x.size(s)))
            elif isinstance(s, (tuple, list)):
                total = 1
                for j in s:
                    total *= int(x.size(j))
                newshape.append(total)
            else:
                raise ValueError(
                    "shape spec must be either int or tuple, got {}".format(s))
        return x.view(*newshape)

    def __repr__(self):
        return "Reshape({})".format(
            ", ".join([repr(x) for x in self.shape]))


class Viewer(nn.Module):
    """Module equivalent of x.view(*args)"""

    def __init__(self, *args):
        super().__init__()
        self.shape = args

    def forward(self, x):
        return x.view(*self.shape)

    def __repr__(self):
        return "Viewer({})".format(
            ", ".join([repr(x) for x in self.shape]))

class LSTM(nn.Module):
    """LSTM wrapper that discards the state."""
    def __init__(self, *args, **kw):
        super().__init__()
        self.lstm = nn.LSTM(*args, **kw)

    def forward(self, *args, **kw):
        return self.lstm(*args, **kw)[0]

class BDL_LSTM(nn.Module):
    """A simple bidirectional LSTM.

    All the sequence processing layers use BDL order by default to
    be consistent with 1D convolutions.
    """

    def __init__(self, ninput=None, noutput=None, num_layers=1, bidirectional=False, batch_first=True):
        super().__init__()
        assert ninput is not None
        assert noutput is not None
        self.lstm = nn.LSTM(ninput, noutput, num_layers=num_layers,
                            bidirectional=bidirectional)

    def forward(self, seq, volatile=False, verbose=False):
        seq = reorder(seq, "BDL", "LBD")
        output, _ = self.lstm(seq)
        return reorder(output, "LBD", "BDL")


class BDHW_LSTM(nn.Module):
    """A 2D LSTM module.

    Input order as for 2D convolutions.
    """

    def __init__(self, ninput=None, noutput=None, nhidden=None, 
                 num_layers=1, bidirectional=True):
        super().__init__()
        nhidden = nhidden or noutput
        ndir = bidirectional+1
        self.hlstm = nn.LSTM(ninput, nhidden, num_layers=num_layers,
                             bidirectional=bidirectional)
        self.vlstm = nn.LSTM(nhidden*ndir, noutput, num_layers=num_layers,
                             bidirectional=bidirectional)

    def forward(self, img):
        b, d, h, w = img.shape
        hin = reorder(img, "BDHW", "WHBD").view(w, h*b, d)
        hout, _ = self.hlstm(hin)
        vin = reorder(hout.view(w, h, b, -1), "WHBD", "HWBD").view(h, w*b, -1)
        vout, _ = self.vlstm(vin)
        return reorder(vout.view(h, w, b, -1), "HWBD", "BDHW")



class BDHW_LSTM_to_BDH(nn.Module):
    """An LSTM that summarizes 2D down to 1D along the last dim."""
    def __init__(self, ninput=None, noutput=None):
        super().__init__()
        assert ninput is not None
        assert noutput is not None
        self.lstm = nn.LSTM(ninput, noutput, 1, bidirectional=False)

    def forward(self, img, volatile=False):
        noutput = self.lstm.hidden_size
        b, d, h, w = img.size()
        seq = reorder(img, "BDHW", "WBHD").view(w, b*h, d)
        out, (_, state) = self.lstm(seq)
        assert state.size() == (1, b*h, noutput), ((w, b*h, noutput), state.size())
        return reorder(state.view(b, h, noutput), "BHD", "BDH")

### Wrap-Around Modules

class NoopSub(nn.Module):
    """Noop wrap-around module (for testing/exploration)."""
    def __init__(self, *args, sub=None, **kw):
        super().__init__()
        self.sub = sub
    def forward(self, x):
        return self.sub(x)

class KeepSize(nn.Module):
    """Run layers, then upsample back to the original.
    """
    def __init__(self, mode="bilinear", sub=None, dims=None):
        super().__init__()
        if isinstance(sub, list):
            sub = nn.Sequential(*sub)
        self.sub = sub
        self.mode = mode
        self.dims = dims
    def forward(self, x):
        y = self.sub(x)
        if self.dims is None:
            size = x.size()[2:]
        else:
            size = [x.size(i) for i in self.dims]
        kw = dict(align_corners=False) if self.mode != "nearest" else {}
        try:
            return F.interpolate(y, size=size, mode=self.mode, **kw)
        except Exception as exn:
            print("error:", x.size(), y.size(), self.dims, size, self.mode, file=sys.stderr)
            raise exn

class Additive(nn.Module):
    """Additive wrap-around module for Resnet-style architectures.

    :args: modules whose output is to be added
    :post: module to execute after everything has been added
    """
    def __init__(self, *args, post=None):
        super().__init__()
        self.sub = nn.ModuleList(args)
        self.post = None
    def forward(self, x):
        y = self.sub[0](x)
        for f in self.sub[1:]:
            y = y + f(x)
        if self.post is not None:
            y = self.post(y)
        return y

class Parallel(nn.Module):
    """Run modules in parallel and concatenate the results."""
    def __init__(self, *args, dim=1):
        super().__init__()
        self.args = args
        for i, arg in enumerate(args):
            if isinstance(arg, list):
                arg = nn.Sequential(*arg)
            self.add_module(str(i), arg)
        self.dim = dim

    def forward(self, x):
        results = [f(x) for f in self.args]
        return torch.cat(results, dim=self.dim)


class SimplePooling2d(nn.Module):
    """Perform max pooling/unpooling"""

    def __init__(self, sub, mp=2, **kw):
        super().__init__()
        self.pool = torch.nn.MaxPool2d(mp, return_indices=True, ceil_mode=True)
        self.sub = nn.Sequential(*sub) if isinstance(sub, list) else sub
        self.unpool = torch.nn.MaxUnpool2d(mp)

    def forward(self, x):
        y, indices = self.pool(x)
        z = self.sub(y)
        indices, z = conform1(indices, z, dims=[0, 2, 3])
        assert z.shape == indices.shape, (z.shape, indices.shape)
        return self.unpool(z, indices)

    def __repr__(self):
        return "Pooling2d(\n"+repr(self.sub)+"\n)"


class AcrossPooling2d(nn.Module):
    """Perform max pooling/unpooling with across accumulation."""

    def __init__(self, sub, across, mp=2, **kw):
        super().__init__()
        self.pool = torch.nn.MaxPool2d(mp, return_indices=True, ceil_mode=True)
        self.sub = nn.Sequential(*sub) if isinstance(sub, list) else sub
        self.across = nn.Sequential(*across) if isinstance(across, list) else across
        self.unpool = torch.nn.MaxUnpool2d(mp)

    def forward(self, x):
        y, indices = self.pool(x)
        z = self.sub(y)
        indices, z = conform1(indices, z, slop=2, dims=[0, 2, 3])
        up = self.unpool(z, indices)
        across = self.across(x)
        up, across = conform(up, across, slop=2, dims=[0, 2, 3])
        return torch.cat([across, up], dim=1)





#
# Copyright (c) 2017-2019 TBD. All rights reserved.
# This file is part of TBD (see TBD).
# See the LICENSE file for licensing terms (TBD).
#

"""A set of helper functions for dealing uniformly with tensors and
ndarrays."""

import numpy as np
import torch
import torch.nn.functional as F
from scipy import ndimage
from torch import autograd, nn, optim
from torch.autograd import Variable

def typeas(a, b):
    return a.type(b.dtype).to(b.device)

def bhwd2bdhw(images, depth1=False):
    images = as_torch(images)
    if depth1:
        assert len(shp(images)) == 3, shp(images)
        images = images.unsqueeze(3)
    assert len(shp(images)) == 4, shp(images)
    return images.permute(0, 3, 1, 2)


def bdhw2bhwd(images, depth1=False):
    images = as_torch(images)
    assert len(shp(images)) == 4, shp(images)
    images = images.permute(0, 2, 3, 1)
    if depth1:
        assert images.size(3) == 1
        images = images.index_select(3, 0)
    return images


def reorder(batch, inp, out):
    """Reorder the dimensions of the batch from inp to out order.

    E.g. BHWD -> BDHW.
    """
    if inp is None:
        return batch
    if out is None:
        return batch
    assert isinstance(inp, str)
    assert isinstance(out, str)
    assert len(inp) == len(out), (inp, out)
    assert rank(batch) == len(inp), (rank(batch), inp)
    result = [inp.find(c) for c in out]
    # print ">>>>>>>>>>>>>>>> reorder", result
    for x in result:
        assert x >= 0, result
    if is_tensor(batch):
        return batch.permute(*result)
    elif isinstance(batch, np.ndarray):
        return batch.transpose(*result)


def assign(dest, src, transpose_on_convert=None):
    """Resizes the destination and copies the source."""
    src = as_torch(src, transpose_on_convert)
    if isinstance(dest, Variable):
        dest.data.resize_(*shp(src)).copy_(src)
    elif isinstance(dest, torch.Tensor):
        dest.resize_(*shp(src)).copy_(src)
    else:
        raise ValueError("{}: unknown type".format(type(dest)))

def sequence_is_normalized(a, dim=-1):
    sums = a.detach().sum(dim)
    err = (sums-1.0).abs().max().item()
    return err < 1e-4
    
def ctc_align(prob, target):
    """Perform CTC alignment on torch sequence batches (using ocrolstm)"""
    import cctc2
    prob_ = prob.cpu()
    target = target.cpu()
    b, l, d = prob.size()
    bt, lt, dt = target.size()
    assert bt == b, (bt, b)
    assert dt == d, (dt, d)
    assert sequence_is_normalized(prob, 2), prob
    assert sequence_is_normalized(target, 2), target
    result = cctc2.ctc_align_targets_batch(prob_, target)
    return typeas(result, prob)


def ctc_loss(probs, target):
    """A CTC loss function for BLD sequence training."""
    assert probs.is_contiguous()
    assert target.is_contiguous()
    assert sequence_is_normalized(probs)
    assert sequence_is_normalized(target)
    aligned = ctc_align(probs, target)
    assert aligned.size() == probs.size(), \
        (aligned.size(), probs.size())
    deltas = aligned - probs
    probs.backward(deltas.contiguous())
    return typeas(deltas, probs), typeas(aligned, probs)


class LearningRateSchedule(object):
    def __init__(self, schedule):
        if ":" in schedule:
            self.learning_rates = [
                [float(y) for y in x.split(",")] for x in schedule.split(":")]
            assert self.learning_rates[0][0] == 0
        else:
            lr0 = float(schedule)
            self.learning_rates = [[0, lr0]]

    def __call__(self, count):
        _, lr = self.learning_rates[0]
        for n, l in self.learning_rates:
            if count < n:
                break
            lr = l
        return lr



#
# Copyright (c) 2017-2019 TBD. All rights reserved.
# This file is part of TBD (see TBD).
# See the LICENSE file for licensing terms (TBD).
#
from __future__ import unicode_literals

import pdb
from builtins import next
import sys
import os
import numpy as np

import torch
from torch import nn
from torchmore import layers
from torchmore import flex
from torchmore import combos

def test_UnetLayer():
    mod = combos.UnetLayer(33, sub=nn.Identity(), post=nn.Conv2d(33+11, 7, 3, padding=1))
    a = torch.ones((17, 11, 64, 64))
    b = mod(a)
    assert tuple(b.size()) == (17, 7, 64, 64)

def test_ResnetBlock():
    mod = combos.ResnetBlock(33)
    a = torch.ones((17, 33, 64, 64))
    b = mod(a)
    assert tuple(b.size()) == (17, 33, 64, 64)

def test_ResnetBottleneck():
    mod = combos.ResnetBottleneck(33, 22)
    a = torch.ones((17, 33, 64, 64))
    b = mod(a)
    assert tuple(b.size()) == (17, 33, 64, 64)



#!/usr/bin/python3
#
# Copyright (c) 2017-2019 TBD. All rights reserved.
# This file is part of TBD (see TBD).
# See the LICENSE file for licensing terms (TBD).
#

import setuptools

VERSION = "0.1.0"

setuptools.setup(
    author="Thomas Breuel",
    author_email="tmbdev+removeme@gmail.com",
    description="Useful additional layers for PyTorch.",
    install_requires="simplejson braceexpand msgpack pyyaml numpy torch Pillow torchvision scipy".split(),
    keywords="object store, client, deep learning",
    license="MIT",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    name='torchmore',
    packages=["torchmore"],
    python_requires=">=3.6",
    # scripts=scripts,
    url="http://github.com/tmbdev/torchmore",
    version=VERSION,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7"
    ],
)




#
# Copyright (c) 2017-2019 TBD. All rights reserved.
# This file is part of TBD (see TBD).
# See the LICENSE file for licensing terms (TBD).
#

import numpy as np
import torch
from torch import autograd, nn
from torch.autograd import Variable
import warnings

from torchmore import flex, layers

kw = dict(kernel_size=3, padding=1)

def test_Linear():
    mod = flex.Linear(17)
    a = torch.zeros((7, 3))
    b = mod(a)
    assert b.size() == (7, 17)
    a = torch.zeros((4, 3))
    b = mod(a)
    assert b.size() == (4, 17)


def test_Conv1d():
    mod = flex.Conv1d(17, 3, padding=1)
    a = torch.zeros((7, 3, 99))
    b = mod(a)
    assert b.size() == (7, 17, 99)
    a = torch.zeros((4, 3, 9))
    b = mod(a)
    assert b.size() == (4, 17, 9)


def test_Conv2d():
    mod = flex.Conv2d(17, 3, padding=1)
    a = torch.zeros((7, 3, 99, 88))
    b = mod(a)
    assert b.size() == (7, 17, 99, 88)
    a = torch.zeros((4, 3, 9, 8))
    b = mod(a)
    assert b.size() == (4, 17, 9, 8)


def test_Conv3d():
    mod = flex.Conv3d(17, 3, padding=1)
    a = torch.zeros((7, 3, 99, 88, 77))
    b = mod(a)
    assert b.size() == (7, 17, 99, 88, 77)
    a = torch.zeros((4, 3, 9, 8, 7))
    b = mod(a)
    assert b.size() == (4, 17, 9, 8, 7)

def test_ConvTranspose1d():
    mod = flex.ConvTranspose1d(17, 3, padding=1)
    a = torch.zeros((7, 3, 99))
    b = mod(a)
    assert b.size() == (7, 17, 99)
    a = torch.zeros((4, 3, 9))
    b = mod(a)
    assert b.size() == (4, 17, 9)


def test_ConvTranspose2d():
    mod = flex.ConvTranspose2d(17, 3, padding=1, stride=2)
    a = torch.zeros((7, 3, 99, 88))
    b = mod(a)
    assert b.size() == (7, 17, 197, 175)
    a = torch.zeros((4, 3, 9, 8))
    b = mod(a)
    assert b.size() == (4, 17, 17, 15)


def test_ConvTranspose3d():
    mod = flex.ConvTranspose3d(17, 3, padding=1)
    a = torch.zeros((7, 3, 99, 88, 77))
    b = mod(a)
    assert b.size() == (7, 17, 99, 88, 77)
    a = torch.zeros((4, 3, 9, 8, 7))
    b = mod(a)
    assert b.size() == (4, 17, 9, 8, 7)


def test_LSTM():
    mod = flex.LSTM(17)
    a = torch.zeros((99, 7, 4))
    b = mod(a)
    assert b.size() == (99, 7, 17)

def test_Lstm1d():
    mod = flex.Lstm1d(17)
    a = torch.zeros((7, 3, 99))
    b = mod(a)
    assert b.size() == (7, 17, 99)
    a = torch.zeros((4, 3, 9))
    b = mod(a)
    assert b.size() == (4, 17, 9)

def test_Lstm2d():
    mod = flex.Lstm2d(17)
    a = torch.zeros((7, 3, 99, 88))
    b = mod(a)
    assert b.size() == (7, 34, 99, 88)
    a = torch.zeros((4, 3, 9, 8))
    b = mod(a)
    assert b.size() == (4, 34, 9, 8)

def test_BatchNorm1d():
    mod = flex.BatchNorm1d(3)
    a = torch.zeros((7, 3, 99))
    b = mod(a)
    assert b.size() == (7, 3, 99)
    a = torch.zeros((4, 3, 9))
    b = mod(a)
    assert b.size() == (4, 3, 9)


def test_BatchNorm2d():
    mod = flex.BatchNorm2d(3)
    a = torch.zeros((7, 3, 99, 88))
    b = mod(a)
    assert b.size() == (7, 3, 99, 88)
    a = torch.zeros((4, 3, 9, 8))
    b = mod(a)
    assert b.size() == (4, 3, 9, 8)


def test_BatchNorm3d():
    mod = flex.BatchNorm3d(3)
    a = torch.zeros((7, 3, 99, 88, 77))
    b = mod(a)
    assert b.size() == (7, 3, 99, 88, 77)
    a = torch.zeros((4, 3, 9, 8, 7))
    b = mod(a)
    assert b.size() == (4, 3, 9, 8, 7)


def test_shape_inference():
    mod = nn.Sequential(flex.Conv1d(3, 3, padding=1))
    print(mod)
    assert "Flex" in repr(mod)
    a = torch.zeros((7, 3, 99))
    b = mod(a)
    assert b.size() == (7, 3, 99)
    assert "Flex" in repr(mod)
    print(mod)
    flex.shape_inference(mod, a.shape)
    print(mod)
    assert "Flex" not in repr(mod)
    a = torch.zeros((4, 3, 9))
    b = mod(a)
    assert b.size() == (4, 3, 9)
